/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 1994-2006, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

    http://www.johncon.com/ntropix/
    http://www.johncon.com/

------------------------------------------------------------------------------

r8.c is for controlling an R. L. Drake R8 Communications Receiver via
an RS232C serial port. The default port is /dev/ttyS0 at a fixed 9600
baud rate. The default semaphore lock file is /var/lock/LCK..ttyS0,
which has an fcntl(2) advisory lock applied, also. The serial port is
controlled through termios(3), and is restored to original values on
exit. If the input is stdin, and a tty, (as opposed to a file, pipe,
named pipe, or redirection,) the terminal attributes are changed to
through termios(3) to non-cannonical, one character at a time, which
is restored to original values on exit.  The serial port child read
process is created by fork(2) from the parent write process, and both
processes respond to SIGINT = ^C and SIGTERM signals.

R. L. Drake R8 RS232C command interface, (from the Drake R8
Communications Receiver Owner's Manual, DOM-385112990B-5-1991, page
20):

R8 COMMAND:                RS232 COMMAND:       R8 RESPONSE

AGC set                    AF,                  LF
                           AS,                  LF
                           AO                   LF

ANT select                 A1,                  LF
                           A2,                  LF
                           AC                   LF

Noise blanker set          BW,                  LF
                           BN,                  LF
                           BF                   LF

Step change                ST (scrolls)         LF

Set frequency              F xxxxxxx            CR LF if OK,
                                                CR if not

Syncro det. set            SO,                  LF
                           SF                   LF

RF select                  G+,                  LF
                           G-,                  LF
                           G0                   LF

Notch set                  NO,                  LF
                           NF                   LF

Mode select                M6 (AM,)             LF
                           M5 (FM,)             LF
                           M4 (CW,)             LF
                           M3 (RTTY,)           LF
                           M2 (LSB,)            LF
                           M1 (USB)             LF

VFO select                 VA,                  LF
                           VB                   LF

Bandwidth select           W0 (0.5,)            LF
                           W1 (1.8,)            LF
                           W2 (2.3,)            LF
                           W4 (4.0,)            LF
                           W6 (6.0)             LF

Memory mode                C                    LF

Memory channel select      Cx (x = 0 to 99)     CR LF if OK,
                                                CR if not

Program Memory channel     PRx (x = 0 to 99)    CR LF if OK,
                                                CR if not

Power on/off               PO,                  LF
                           PF                   LF

Timer relay on/off         TO,                  LF
                           TF                   LF

Identify receiver          ID                   R8 CR LF

Tune up (freq. or scan)    U                    None

Tune down ("   "  ")       D                    None

Scan select                Sx (x = 1-3, 4-5)    LF

Scan start/stop            SC (toggle)          LF

Report frequency           RF                   xxx.xxxxx mHz CR LF

Report frequency,          RFR                  xxx.xxxxx mHz CR LF
repeat every 5 seconds

Report mode                RM                   xxxxx CR LF

Report mode,               RMR                  xxxxx CR LF
repeat every 5 seconds

Report mem channel         RC                   xx CR LF

Report mem channel,        RCR                  xx CR LF
repeat every 5 seconds

Report all                 RA                   nn aaaa dd.ddddd mHz

Report all,                RAR                  nn aaaa dd.ddddd mHz
repeat every 5 seconds

Report stop                RS                   LF

All commands are terminated with a CR.

Note: The Linux minicom(1) communications program can also be used to
control the R8:

    1) Ctrl-A, Z for help screen
    2) P for comm parameters
    3) EMRUW, 9600, even parity, 7E1, 7 data bits, 1 stop bit
    4) Ctrl-A, Z for help screen
    5) E for echo

There is no echo from the Drake R8, and commands must be in upper case.

The command set, and returns, used in this program is identical to the
command set outlined in R. L. Drake R8 RS232C command interface, (from
the Drake R8 Communications Receiver Owner's Manual,
DOM-385112990B-5-1991, page 20,) except that 'H' was added to print a
help screen, 'Q' and 'X' were added to exit the program.

As an additional expediency, the following frequency translation
occurs, (if the first character of a command word is a number):

    0010000-0099999 -> F0010000-F0099999
    0100000-0999999 -> F0100000-F0999999
    1000000-3000000 -> F1000000-F3000000

    400000-999999 -> F0400000-F0999999

    First character of a word a number?

        First character a 0, 1, 2, or, 3?

            Add a leading 'F' character and output the character

        First character a 4, 5, 6, 7, 8, or, 9?

            Add a leading 'F', follwed by a '0', and output the
            character

$Revision: 0.0 $
$Date: 2006/06/18 19:36:00 $
$Id: r8.c,v 0.0 2006/06/18 19:36:00 john Exp $
$Log: r8.c,v $
Revision 0.0  2006/06/18 19:36:00  john
Initial revision


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <ctype.h>
#include <errno.h>

#define _POSIX_SOURCE 1 /* POSIX compliant source */

#define BAUDRATE B9600 /* baud rate for serial port communication */
#define SPDEVICE "/dev/ttyS0" /* default serial port device */
#define LOCKFILE "/var/lock/LCK..ttyS0" /* default serial port lock file name */

#define BUFLEN BUFSIZ /* size of record buffers */

static char rcsid[] = "$Id: r8.c,v 0.0 2006/06/18 19:36:00 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 1994-2006, John Conover, All Rights Reserved"; /* the copyright banner */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Control a R. L. Drake R8 Communications Receiver via an RS232C serial port\n",
    "Usage: r8 [-d device] [-i infilename] [-o outfilename] [-l lockfilename] [-v]\n",
    "    -d device, device name, (/dev/ttyS0)\n",
    "    -i infilename, input filename, (stdin)\n",
    "    -l lockfilename, lock file name, (/var/lock/LCK..ttyS0)\n",
    "    -o outfilename, output file name, (stdout)\n",
    "    -v, print the program's version information\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error\n",
    "Error in program argument(s)\n",
    "Error opening serial device\n",
    "Error getting serial device attributes\n",
    "Error installing signal handler\n",
    "Error flushing serial device attributes\n",
    "Error setting serial device attributes\n",
    "Error writing to serial device\n",
    "Error resetting serial device attributes\n",
    "Error closing serial device\n",
    "Error opening input file\n",
    "Error closing input file\n",
    "Error creating lock file\n",
    "Error removing lock file\n",
    "Error forking to child read process\n",
    "Error reading from serial device\n",
    "Error locking lock file\n",
    "Error writing to lock file\n",
    "Error opening output file\n",
    "Error closing output file\n",
    "Error reading input file\n",
    "Error writing to output file\n",
    "Error enabling blocking on serial device\n",
    "Error flushing output file\n",
    "Error setting baud rate\n",
    "Error flushing serial device\n",
    "Error read buffer overflow\n",
    "Error write buffer overflow\n",
    "Error getting tty attributes\n",
    "Error setting tty attributes\n",
    "Error resetting tty attributes\n",
    "Error flushing tty attributes\n"
};

#define NOERROR 0 /* negative error values, one for each index in the error message array */
#define EARGS -1
#define EOPEN -2
#define EGATTR -3
#define ESIG -4
#define EFLUSH -5
#define ESATTR -6
#define EWRITE -7
#define ERATTR -8
#define ECLOSE -9
#define EIOPEN -10
#define EICLOSE -11
#define ELOCKF -12
#define ELOCKR -13
#define EFORK -14
#define EREAD -15
#define EFCNL -16
#define EWLCK -17
#define EOOPEN -18
#define EOCLOSE -19
#define ERDIN -20
#define EOWRITE -21
#define EBLOCK -22
#define EOFLUSH -23
#define EBAUD -24
#define EDFLUSH -25
#define EREOVFL -26
#define EWROVFL -27
#define EGTTYR -28
#define ESTTYR -29
#define ERTTYR -30
#define EFLUSHTTY -31
#define EXIT -32 /* used as a return code to exit */

#ifdef __STDC__

static void print_message (int pretval); /* print any error messages */
static void spclose (int err); /* interrupt handler */
static int child_read (int fdd, int infiled); /* child read loop, process all read records from the r8 */
static int parent_write (int fdd, int infiled, int outfiled); /* parent write loop, process all write records to the r8 */
static int interpreter (int c, char *buf); /* interpreter for r8 commands */

#else

static void print_message (); /* print any error messages */
static void spclose (); /* interrupt handler */
static int child_read (); /* child read loop, process all read records from the r8 */
static int parent_write (); /* parent write loop, process all write records to the r8 */
static int interpreter (); /* interpreter for r8 commands */

#endif

static char serial_lockname[] = {LOCKFILE}, /* default lock file name */
            *lockname = serial_lockname, /* default lock file name */
            serial_device[] = {SPDEVICE}; /* default serial port device name */

static int retval = NOERROR, /* return value, assume no error */
           fd = -1, /* serial port device file descriptor */
           infile = STDIN_FILENO, /* input file descriptor */
           outfile = STDOUT_FILENO, /* output file file descriptor */
           semaphorelock = -1, /* lock file file descriptor */
           state = 0, /* state of serial port setup, 0 = not set up, 1 = original device attributes successfully saved */
           tty_set = 0, /* state of tty setup, 0 = not set up, 1 = original tty attributes successfully saved */
           remove_lockfile = 0; /* remove lock file, 0 = no, 1 = yes */

static pid_t pid_write = (pid_t) 0, /* pid of parent write process */
             pid_read = (pid_t) 0; /* pid of child read process */

static struct termios oldtio, /* original device attributes, restored on exit */
                      oldtty; /* original terminal attributes, restored on exit */

static struct flock lockfile; /* lock file structure */

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    static char pid_str[BUFLEN]; /* pid string */

    char c, /* command line switch */
         *device = serial_device, /* default serial port device name */
         *in_name = (char *) 0, /* input file name */
         *out_name = (char *) 0; /* output file name */

    struct termios newtio, /* device attributes used by this program */
                   newtty; /* new terminal attributes */

    pid_write = getpid (); /* save the pid of parent write process */

    while ((c = (char) getopt (argc, argv, "d:i:l:o:v")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'd': /* request for device name? */

                device = optarg; /* yes, save the device name */
                break;

            case 'i': /* request for input file name? */

                in_name = optarg; /* yes, save the input file name */
                break;

            case 'l': /* request for lock file name? */

                lockname = optarg; /* yes, save the lock file name */
                break;

            case 'o': /* request for output file name? */

                out_name = optarg; /* yes, save the output file name */
                break;

            case 'v':

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if (retval == NOERROR) /* enough arguments? */
    {
        retval = ESIG; /* assume error installing signal handler */

        if (signal (SIGINT, spclose) != SIG_ERR) /* install the interrupt handler for SIGINT = ^C */
        {

            if (signal (SIGTERM, spclose) != SIG_ERR) /* install the interrupt handler for SIGTERM */
            {

                if (signal (SIGCHLD, spclose) != SIG_ERR) /* install the interrupt handler for SIGCHLD */
                {
                    lockfile.l_type = F_WRLCK; /* exclusive write lock the entire file */
                    lockfile.l_whence = (short int) SEEK_SET;
                    lockfile.l_start = (off_t) 0;
                    lockfile.l_len = (off_t) 0;

                    retval = ELOCKF; /* assume error creating lock file */

                    if ((semaphorelock = open (lockname, O_RDWR | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH)) >= 0) /* open the lock file exclusively */
                    {
                        remove_lockfile = 1; /* set the remove lock file, 0 = no, 1 = yes */
                        retval = NOERROR; /* assume no error */
                    }

                    else if ((semaphorelock = open (lockname, O_RDWR, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH)) >= 0) /* open the lock file if it exists */
                    {
                        retval = NOERROR; /* assume no error */
                    }

                    if (retval == NOERROR) /* any errors? */
                    {
                        retval = EFCNL; /* assume error locking lock file */

                        if (fcntl (semaphorelock, F_SETLK, &lockfile) == 0) /* exclusive write lock the lock file */
                        {
                            remove_lockfile = 1; /* set the remove lock file, 0 = no, 1 = yes */
                            retval = EWLCK; /* assume error writing to lock file */

                            if (ftruncate (semaphorelock, 0) == 0) /* truncate the lock file */
                            {
                                (void) sprintf (pid_str, "%10d\n", pid_write); /* the pid string is the PID right justified to 10 characters plus a '\n' EOL */

                                if (write (semaphorelock, pid_str, strlen (pid_str)) >= 0) /* write the serial port dialer string to the serial port */
                                {
                                    retval = EOPEN; /* assume error opening device */

                                    if ((fd = open (device, O_RDWR | O_NOCTTY | O_NDELAY)) >= 0) /* open the serial port device file descriptor */
                                    {
                                        retval = EBLOCK; /* assume error enabling blocking on serial device */

                                        if (fcntl (fd, F_SETFL, 0) == 0) /* enable read blocking, i.e., read () waits until a character is in the buffer */
                                        {
                                            retval = EFLUSH; /* assume error flushing device attributes */

                                            if (tcflush (fd, TCIOFLUSH) == 0) /* flush device input/output */
                                            {
                                                retval = EGATTR; /* assume error getting device attributes */

                                                if (tcgetattr (fd, &oldtio) == 0) /* save original port settings */
                                                {
                                                    state = 1; /* state of serial port setup, 0 = not set up, 1 = original device attributes successfully saved */
                                                    memcpy (&newtio, &oldtio, sizeof (newtio)); /* copy the original device attributes */
                                                    retval = EBAUD; /* assume error setting baud rate */

                                                    if (cfsetispeed (&newtio, BAUDRATE) == 0) /* set the baud rate */
                                                    {

                                                        if (cfsetospeed (&newtio, BAUDRATE) == 0)
                                                        {

                                                            /* terminal set up */

                                                            newtio.c_cflag |= (CLOCAL | CREAD); /* local mode, enable receiver */
                                                            newtio.c_cflag |= PARENB; /* even parity */
                                                            newtio.c_cflag &= ~PARODD;
                                                            newtio.c_cflag &= ~CSTOPB;
                                                            newtio.c_cflag &= ~CSIZE; /* 7 data bits */
                                                            newtio.c_cflag |= CS7;
                                                            newtio.c_cflag |= CRTSCTS; /* HW flow control */
                                                            newtio.c_lflag &= ~(ICANON | ECHO | ISIG); /* non-canonical raw input with echo */

                                                            retval = ESATTR; /* assume error setting serial device attributes */

                                                            if (tcsetattr (fd, TCSANOW, &newtio) == 0) /* set the device attributes */
                                                            {
                                                                retval = EFLUSH; /* assume error flushing device attributes */

                                                                if (tcflush (fd, TCIOFLUSH) == 0) /* flush device input/output */
                                                                {
                                                                    retval = EOOPEN; /* assume error opening output file */

                                                                    if ((outfile = (out_name == (char *) 0) ? outfile : open (out_name, O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH )) >= 0) /* open the output file */
                                                                    {
                                                                        retval = EIOPEN; /* assume error opening file */

                                                                        if ((infile = (in_name == (char *) 0) ? infile : open (in_name, O_RDONLY | O_NOCTTY)) >= 0) /* open the input file */
                                                                        {
                                                                            retval = NOERROR; /* assume no error */

                                                                            if (infile == STDIN_FILENO) /* input file descriptor stdin? */
                                                                            {
                                                                                retval = EFLUSHTTY; /* assume error flushing tty attributes */

                                                                                if (tcflush (infile, TCIOFLUSH) == 0) /* flush tty input/output */
                                                                                {
                                                                                    retval = EGTTYR; /* assume error getting tty attributes */

                                                                                    if (tcgetattr (infile, &oldtty) >= 0) /* attempt to save the terminal settings; will fail if stdin is a pipe or redirection */
                                                                                    {
                                                                                       tty_set = 1; /* state of tty setup, 0 = not set up, 1 = original tty attributes successfully saved */
                                                                                       memcpy (&newtty, &oldtty, sizeof (newtty)); /* copy the original tty attributes */

                                                                                       newtty.c_lflag  &= ~(ECHO | ICANON); /* echo off, cannonical mode off */
                                                                                       newtty.c_cc[VMIN] = 1; /* one byte at a time */
                                                                                       newtty.c_cc[VTIME] = 0; /* no timer */

                                                                                       retval = ESTTYR; /* assume error setting tty attributes */

                                                                                       if (tcsetattr (infile, TCSAFLUSH, &newtty) >= 0) /* set the tty attributes */
                                                                                       {
                                                                                           retval = EFLUSHTTY; /* assume error flushing tty attributes */

                                                                                           if (tcflush (infile, TCIOFLUSH) == 0) /* flush tty input/output */
                                                                                           {
                                                                                               retval = NOERROR; /* assume no error */
                                                                                           }

                                                                                       }

                                                                                    }

                                                                                }

                                                                                else
                                                                                {

                                                                                    if ((errno == ENOTTY) || (errno == EINVAL)) /* ENOTTY is a redirected stdin, EINVAL is a piped stdin; stdin a pipe/redirect? */
                                                                                    {
                                                                                        retval = NOERROR; /* yes, assume no error */
                                                                                    }

                                                                                }

                                                                            }

                                                                            if (retval == NOERROR) /* any errors? */
                                                                            {

                                                                                if ((pid_read = fork ()) < 0) /* fork the child read process */
                                                                                {
                                                                                    retval = EFORK; /* assume error forking to child read process */
                                                                                }

                                                                                else if (pid_read == 0) /* child or parent process, (pid_read is zero in the child process, the child process' PID in the parent)? */
                                                                                {

                                                                                    /* start of child read process; pid_write still contains the PID of the parent write process */

                                                                                    retval = child_read (fd, outfile); /* child read loop, process all read records from the r8 */

                                                                                    /* end of the child read process, fall through and exit */
                                                                                }

                                                                                else /* wasn't child process, must be parent */
                                                                                {

                                                                                    /* start of the parent write process; pid_read contains the PID of the child read process */

                                                                                    retval = parent_write (fd, infile, outfile); /* parent write loop, process all write records to the r8 */

                                                                                    /* end of the parent write process, fall through and exit */

                                                                                }

                                                                            }

                                                                        }

                                                                    }

                                                                }

                                                            }

                                                        }

                                                    }

                                                }

                                            }

                                        }

                                    }

                                }

                            }

                        }

                    }

                }

            }

        }

    }

    spclose (0); /* shutdown */
    exit (-retval); /* exit with the error value, formality: spclose () did the exit */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int pretval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer pretval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer pretval:

        1) If pretval is zero, print nothing-normal/successful program
           exit.

        2) Else if pretval is unity, print help.

        3) Else pretval is not zero or unity, it is an error code, print
           the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int pretval)

#else

static void print_message (pretval)
int pretval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (pretval) /* which return value */
    {

        case 0: /* program ended without errors, print nothing */

            break;

        case 1: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[pretval]);
            break;

    }

}

/*

Interrupt handler and shutdown.

static void spclose (int err);

I) Prerequisites:

    A) All file opening and serial port initialization is done by the
       parent write process, (excepting the log file, which opened and
       closed dynamically by the child read process-and is normally
       closed,)

        1) The parent write process should close all files, (except
           the log file,) and restore the serial port.

        2) The child read process should close the log file.

    B) Called with SIGINT = ^C, or, SIGTERM, or, SIGCHLD interrupts,
       or, with a negative argument representing an internal error,
       or, an argument of zero for no errors or interrupts.

    C) There are "edge" issues, (i.e., a flag is tested to see if
       something needs shutting down, and the something is shut down,
       but the program is interrupted-perhaps by SIGINT = ^C
       bashing-before the flag gets reset,) There are no error message
       printed in the shut down section of this routine-everything is
       shut down, and at the end of this routine, the flags are
       queried for a successful shutdown, and shutdown errors, if any,
       printed. The idea being that if an "edge" issue occurred, the
       shut down would occur twice, (presumably giving an error the
       second time, which would not be printed, but as long as the
       flag is reset, no error would be printed.)

    D) The kill of the child read by the parent write process, (and
       vice verse,) is complex:

        1) static passed = reset
        2) if parent
            a) if err is SIGCHLD
                i) wait for the child read process to die
            b) if passed = reset
                i) reset all things in the parent process
                ii) print any errors
                iii) set passed
                iv) if err not SIGCHLD and child read process active,
                    kill child read process
        3) else if child
            a) if passed = reset
                i) close the log file
                ii) print any errors
                iii) set passed
                iv) kill parent
        4) restore all signals

II) Data structures:

    A) if the pid is equal to pid_write then the parent write process
       is executing this routine, else, it is the child read process.

    B) The parent write and child read process have a separate copy of
       the int passed flag-which means a complete pass has been made
       through this routine for the client or parent.

III) Function execution:

    A) Upon entry:

        1) If the PID of the process is the parent write process:

            a) Shut down all devices opened by the parent write
               process.

            b) Print any pending errors that occurred before the
               program entered this routine.

            c) Print any errors after the shutdown is complete.

            d) If the call to this routine was not caused by a dead
               child read process, and the child process is actve,
               kill the child read process.

        2) If the PID of the process is the child read process:

            a) Shut down all devices opened by the child read process.

            b) Print any pending errors that occurred before the
               program entered this routine.

            c) Print any errors after the shutdown is complete.

            d) Kill the parent write process.

        3) Restore the signal handlers for SIGINT = ^C and SIGTERM.

There is no return from this routine. The routine exits with a value
of zero if no errors or interrupts occurred, or, the value of the
interrupt, (a positive number,) if an interrupt occurred, or, with a
negative error number if an an error occurred.

*/

#ifdef __STDC__

static void spclose (int err)

#else

static void spclose (err)
int err;

#endif

{
    static int passed = 0; /* passed this routine, 0 = no, 1 = yes, used by both parent write and child read processes, but are different copies */

    int exit_value,
        local_retval = NOERROR; /* assume no error shutting things down */

    if (pid_write == getpid ()) /* parent write process? */
    {

        if (err == SIGCHLD) /* child read process died? */
        {
            (void) wait ((int *) 0); /* free the child read process */
        }

        if (passed == 0) /* passed this routine, 0 = no, 1 = yes? */
        {

            if (state == 1) /* state of serial port setup, 0 = not set up, 1 = original device attributes successfully saved */
            {
                (void) tcflush (fd, TCIOFLUSH); /* flush device input/output */

                if (tcsetattr (fd, TCSANOW, &oldtio) == 0) /* reset the device attributes */
                {
                    state = 0; /* state of serial port setup, 0 = not set up, 1 = original device attributes successfully saved */
                    (void) tcflush (fd, TCIOFLUSH); /* flush device input/output */
                }

            }

            if (fd > 0) /* serial port device file descriptor successfully opened? */
            {
                (void) tcflush (fd, TCIOFLUSH); /* flush device input/output */

                if (close (fd) == 0) /* yes, close the device */
                {
                    fd = -1; /* reset the serial port device file descriptor */
                }

            }

            if (tty_set == 1) /* state of tty setup, 0 = not set up, 1 = original tty attributes successfully saved */
            {
                (void) tcflush (infile, TCIOFLUSH); /* flush tty input/output */

                if (tcsetattr (infile, TCSANOW, &oldtty) == 0) /* reset the tty attributes */
                {
                    tty_set = 0; /* state of tty setup, 0 = not set up, 1 = original tty attributes successfully saved */
                    (void) tcflush (infile, TCIOFLUSH); /* flush tty input/output */
                }

            }

            if (infile != STDIN_FILENO) /* input file open? */
            {

                if (close (infile) == 0) /* close the input file */
                {
                    infile = STDIN_FILENO; /* reset the input file descriptor */
                }

            }

            if (outfile != STDOUT_FILENO) /* output file open? */
            {

                if (close (outfile) == 0) /* close the output file */
                {
                    outfile = STDOUT_FILENO; /* reset the output file descriptor */
                }

            }

            if (semaphorelock >= 0) /* lock file open? */
            {
                lockfile.l_type = F_UNLCK; /* unlock the file */
                (void) fcntl (semaphorelock, F_SETLKW, &lockfile); /* unlock the lock file */

                if (close (semaphorelock) == 0) /* close the lock file */
                {
                    semaphorelock = -1; /* reset the lock file file descriptor */

                    if (remove_lockfile == 1) /* remove lock file, 0 = no, 1 = yes, set? */
                    {

                        if (unlink (lockname) == 0) /* remove the lock file */
                        {
                            remove_lockfile = 0; /* remove lock file, 0 = no, 1 = yes */
                        }

                    }

                }

            }

            print_message (-retval); /* print any error messages from the program, (there can only be zero or one) */

            /* print any errors shutting things down */

            if (state != 0) /* state of serial port still setup? */
            {print_message (-ERATTR); local_retval = ERATTR;} /* print the error message, set the local error shutting things down */

            if (semaphorelock != -1) /* lock file open? */
            {print_message (-ELOCKR); local_retval = ELOCKR;} /* print the error message, set the local error shutting things down */

            /* probably don't care about these-exit () will handle them */

            if (fd != -1) /* serial port device file descriptor still open? */
            {print_message (-ECLOSE); local_retval = ECLOSE;} /* print the error message, set the local error shutting things down */

            if (tty_set != 0) /* state of tty still setup? */
            {print_message (-ERTTYR); local_retval = ERTTYR;} /* print the error message, set the local error shutting things down */

            if (infile != STDIN_FILENO) /* input file open? */
            {print_message (-EICLOSE); local_retval = EICLOSE;} /* print the error message, set the local error shutting things down */

            if (outfile != STDOUT_FILENO) /* output file open? */
            {print_message (-EOCLOSE); local_retval = EOCLOSE;} /* print the error message, set the local error shutting things down */

            passed = 1; /* passed this routine, 0 = no, 1 = yes, used by both parent write and child read processes, but are different copies */

            if (err != SIGCHLD) /* child read process died? */
            {

                if (pid_read != (pid_t) 0) /* child read process forked? */
                {
                    (void) kill (pid_read, SIGTERM); /* kill the child read process */
                }

            }

        }

    }

    else /* no, child read process */
    {

        if (passed == 0) /* passed this routine, 0 = no, 1 = yes? */
        {
            print_message (-retval); /* print any error messages from the program, (there can only be zero or one) */
            passed = 1; /* passed this routine, 0 = no, 1 = yes, used by both parent write and child read processes, but are different copies */

            if (pid_write != (pid_t) 0) /* parent write process active? */
            {
                (void) kill (pid_write, SIGTERM); /* kill the parent write process */
            }

        }

    }

    /* probably don't care about these-exit () will handle them */

    (void) signal (SIGCHLD, SIG_DFL); /* restore default interrupt for SIGCHLD */
    (void) signal (SIGTERM, SIG_DFL); /* restore default interrupt for SIGTERM */
    (void) signal (SIGINT, SIG_DFL); /* restore default for SIGINT = ^C interrupt */

    exit_value = err; /* assume routine called externally, i.e., SIGINT == ^C or SIGTERM */

    if (exit_value == 0) /* no? */
    {
        exit_value = retval; /* assume error in main body of program, before this routine */

        if (retval == 0) /* no? */
        {
            exit_value = local_retval; /* assume error shutting things down */
        }

    }

    exit (exit_value); /* exit with the value of the interrupt, the value of an error in the main body of this program, the value of an error in this routine, or zero for no error */
}

/*

Parent write loop, process all records from the input file, (usually,
stdin,) to the serial device.

static int parent_write (int fdd, int infiled, int outfiled);

I) Data structure:

    A) The serial port device file descriptor, fdd, should be open, and
       available.

    B) The input file descriptor, infiled, should be open and
       available; it is probably stdin, (but can be a named pipe,
       etc.)

    C) The output file descriptor, outfiled, should be open and
       available; it is probably stdout, (but can be a pipe, etc.)

II) Function execution:

    A) This routine is a "spin loop" that waits for a signal, or, an
       error to terminate the program, while reading characters from
       the input file descriptor, infiled.

    B) Each character from the input file is interpreted by a call to
       the interpret () function.

    C) Interpreted input characters from the input file are printed,
       sequentially, to the r8, and the output file descriptor,
       outfiled.

III) Notes:

    A) Interactive mode, (i.e., infiled is stdin, STDIN_FILENO,)
       requires raw non-cannonical mode since the 'U' and 'D' commands
       do not have line feeds/carriage returns, i.e., processing must
       be done on a character basis. The 'U' and 'D' commands will not
       be available when operating in "batch" mode, either as an input
       file, or piped/redirected stdin.

    B) The r8 requires serial communication a character at a time,
       i.e., output a character, flush the serial buffer, then the
       next character, and so on.

    C) The keyboard/input file echo is done in this function as a
       processing expediency.

Returns any errors encountered, NOERROR on success.

*/

#ifdef __STDC__

static int parent_write (int fdd, int infiled, int outfiled)

#else

static int parent_write (fdd, infiled, outfiled)
int fdd;
int infiled;
int outfiled;

#endif

{
    char buf[BUFSIZ], /* output buffer from command interpreter */
        *ref; /* reference to character in output buffer from command interpreter */

    int local_retval = NOERROR, /* local return value, (so ERDIN = "Error reading input file" will not go to spclose () on a SIGINT = ^C or SIGTERM "edge" in this routine,) assume no error */
        ret, /* return value from read() to check for errors */
        c; /* input character */

    while (local_retval == NOERROR) /* while no error, spin though getting charcters from the keyboard, (actually, waiting for a SIGINT = ^C or SIGTERM) */
    {
        local_retval = ERDIN; /* assume error reading input file */

        if ((ret = read (infiled, &c, 1)) == 1) /* read a character from the input file */
        {
            local_retval = NOERROR; /* assume no error */

            if (interpreter (c, buf) == NOERROR) /* interpret the character, request for exit? */
            {
                ref = buf; /* reference the first character in the output buffer from command interpreter */

                while (*ref != '\0') /* while characters in the output buffer from command interpreter */
                {
                    local_retval = EWRITE; /* assume error writing to serial device */

                    if (write (fdd, ref, (size_t) 1) == 1) /* write/echo the character to the serial device */
                    {
                        local_retval = EDFLUSH; /* assume error flushing serial device */

                        if (tcdrain (fdd) == 0) /* flush the serial device */
                        {
                            local_retval = NOERROR; /* assume no error */

                            if (*ref != '\r' && *ref != '\n') /* do not output CR or LF, CR and LF is the return code from the r8 */
                            {
                                local_retval = EOWRITE; /* assume error writing to output file */

                                if (write (outfiled, ref, (size_t) 1) == 1) /* write the character to the output file */
                                {
                                    local_retval = NOERROR; /* assume no error */
                                }

                            }

                        }

                    }

                    ref ++; /* next character in the output buffer from command interpreter */
                }

            }

            else
            {
                break; /* exit */
            }

        }

        else
        {

            if (ret == 0) /* EOF? */
            {
                local_retval = NOERROR; /* yes, assume no error */
                break; /* and quit with no errors */
            }

        }

    }

    return (retval = local_retval); /* return any errors */
}

/*

Child read loop, process all records from the r8.

static int child_read (int fdd, int outfiled);

I) Data structure:

    A) The serial port device file descriptor, fdd, should be open, and
       available.

    B) The output file descriptor, outfiled, should be open and
       available; it is probably stdout, (but can be a pipe, etc.)

II) Function execution:

    A) This routine is a "spin loop" that waits for a signal, or, an
       error to terminate the program, while reading characters from
       the serial device file descriptor, fdd.

    B) Input characters from the serial device are written to the
       output file descriptor, outfiled.

III) Notes:

    A) The output file structure, (whether stdout, a pipe/redirection,
       or file,) is record oriented as a processing expediency.

    B) It is an error for an EOF from the serial device; the serial
       device should be shutdown from the keyboard/input file, (i.e.,
       it is not necessary to detect a zero byte count from read(2)
       for an EOF condition.)

    C) Buffer overflow is handled by making the buffer size much
       larger than the maximum output string length from the r8,
       (which is a response to a "Report all" command, "*nn aaaa
       dd.ddddd#mHz" with CR LF termination characters,) and if the
       number of characters returned in the buffer is equal to the
       buffer size, issue a buffer overflow error. The r8 does not use
       canonical mode, and frequently a string is returned that does
       not have a termination character sequence, (which could be used
       to check for buffer overflow.)

Returns any errors encountered, NOERROR on success.

*/

#ifdef __STDC__

static int child_read (int fdd, int outfiled)

#else

static int child_read (fdd, outfiled)
int fdd;
int outfiled;

#endif

{
    static char buffer[BUFLEN]; /* input read buffer */

    int local_retval = NOERROR; /* local return value, (so EREAD = "Error reading from device" will not go to spclose () on a SIGINT = ^C or SIGTERM "edge" in this routine,) assume no error */

    ssize_t cnt; /* count of characters read */

    while (local_retval == NOERROR) /* while no error, spin though getting charcters from the serial device, (actually, waiting for a SIGINT = ^C or SIGTERM) */
    {
        local_retval = EREAD; /* assume error reading from device */

        if ((cnt = read (fdd, buffer, BUFLEN)) > 0) /* read characters from the serial device */
        {
            local_retval = EREOVFL; /* assume error read buffer overflow */

            if (cnt <= BUFLEN) /* buffer overflow? */
            {
                local_retval = EOWRITE; /* assume error writing to output file */

                if (write (outfiled, buffer, (size_t) cnt) > 0) /* write the characters to the output file */
                {
                    local_retval = NOERROR; /* assume no error */
                }

            }

        }

    }

    return (retval = local_retval); /* return any errors */
}


/*

Interpreter for r8 commands.

static int interpreter (char c, char *buf);

I) Data structure:

    A) The output buffer, buf, should be allocated.

II) Function execution:

    A) This routine is called from write_parent () to interpret a
       character in a command to the r8:

       1) Characters used:

           a) 'A'-'D', 'E'-'G', 'I', 'M'-'P', 'R'-'W', '0'-'9', +, -,
              '\r', '\n'

       2) Characters never used:

           a) 'H', 'J'-'L', 'Q', 'X'-'Z'

           b) 'H' has been commandeered as a request for help.

           c) 'Q' and 'X' have been commandeered as a request to exit.

       3) First character never used:

           a) 'E', 'H', 'J'-'L', 'O', 'Q', 'X'-'Z', '0'-'9'

           b) '0'-'9' have been commandeered as a request for
              frequency, and follow the following construct:

               0010000-0099999 -> F0010000-F0099999
               0100000-0999999 -> F0100000-F0999999
               1000000-3000000 -> F1000000-F3000000

               400000-999999 -> F0400000-F0999999

               First character of a word a number?

                   First character a 0, 1, 2, or, 3?

                       Add a leading 'F' character and output the
                       character

                   First character a 4, 5, 6, 7, 8, or, 9?

                       Add a leading 'F', follwed by a '0', and output
                       the character

III) Notes:

    A) Buffer overflow is handled by making the buffer size much
       larger than the maximum input string length to the r8, (which
       is "F xxxxxxxx,) which is defined in the parent_write ()
       function, and passed as an argument to this function.

Returns NOERROR on success, the pseudo error EXIT if a request to
exit.

*/

#ifdef __STDC__

static int interpreter (int c, char *buf)

#else

static int interpreter (c, buf)
int c;
char *buf;

#endif

{
    static int cnt = 0; /* count of characters in command word */

    char *ref = buf; /* reference to next character in output buffer */

    int local_retval = NOERROR, /* local return value, assume no error */
        inchar; /* input character */

    inchar = (int) toupper ((unsigned char) c); /* make sure the character is uppercase */

    *ref = '\0'; /* null the output buffer */

    switch (inchar) /* which character? */
    {

        case 'A': /* valid r8 command character? */
        case 'B':
        case 'C':
        case 'E':
        case 'F':
        case 'G':
        case 'I':
        case 'M':
        case 'N':
        case 'O':
        case 'P':
        case 'R':
        case 'S':
        case 'T':
        case 'V':
        case 'W':
        case '+':
        case '-':

            *ref ++ = (char) inchar; /* save the character in the output buffer */
            *ref = '\0'; /* terminate the string */
            cnt ++; /* one more character in the output buffer */
            break;

        case 'D':
        case 'U':

            *ref ++ = (char) inchar; /* save the character in the output buffer */
            *ref = '\0'; /* terminate the string */
            cnt = 0; /* new command word */
            break;

        case 'H': /* request for help? */

            (void) fprintf (stderr, "AGC: AF, AS, AO\n");
            (void) fprintf (stderr, "ANT: A1, A2, AC\n");
            (void) fprintf (stderr, "Bandwidth: W0 (0.5,) W1 (1.8,) W2 (2.3,) W4 (4.0,) W6 (6.0)\n");
            (void) fprintf (stderr, "Frequency: F nnnnnnn\n");
            (void) fprintf (stderr, "Identify receiver: ID\n");
            (void) fprintf (stderr, "Memory channel: C0-C99\n");
            (void) fprintf (stderr, "Memory: C\n");
            (void) fprintf (stderr, "Mode: M6 (AM,) M5 (FM,) M4 (CW,) M3 (RTTY,) M2 (LSB,) M1 (USB)\n");
            (void) fprintf (stderr, "Noise blanker: BW, BN, BF\n");
            (void) fprintf (stderr, "Notch: NO, NF\n");
            (void) fprintf (stderr, "Power on/off: PO, PF\n");
            (void) fprintf (stderr, "Program memory channel: PR0-PR99\n");
            (void) fprintf (stderr, "RF: G+, G-, G0\n");
            (void) fprintf (stderr, "Report all: RA, (repeat = RAR)\n");
            (void) fprintf (stderr, "Report frequency: RF, (repeat = RFR)\n");
            (void) fprintf (stderr, "Report memory channel: RC, (repeat = RCR)\n");
            (void) fprintf (stderr, "Report mode: RM, (repeat = RMR)\n");
            (void) fprintf (stderr, "Report stop: RS\n");
            (void) fprintf (stderr, "Scan start/stop: SC\n");
            (void) fprintf (stderr, "Scan: S1, S2, S3, S4, S5\n");
            (void) fprintf (stderr, "Step: ST\n");
            (void) fprintf (stderr, "Syncro detector: SO, SF\n");
            (void) fprintf (stderr, "Timer relay on/off: TO, TF\n");
            (void) fprintf (stderr, "Tune down: D\n");
            (void) fprintf (stderr, "Tune up: U\n");
            (void) fprintf (stderr, "VFO: VA, VB\n");

            cnt = 0; /* new command word */
            break;

        case 'Q': /* request to exit? */
        case 'X':

            cnt = 0; /* new command word */
            local_retval = EXIT; /* return code to exit */
            break;

        case '0': /* number? */
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':

            /*

            if the first character of a command is a number, it is
            assumed to be a command for frequency, output an 'F'
            followed by:

                if the first character is '4', '5', '6', '7', '8', or,
                '9', then output a leading '0', then the number,
                (i.e., "9500" would be tranlated to "F09500")

                else, output the number, (i.e., "2000" would be
                translated to "F2000")

            */

            if (cnt == 0) /* begining of a command word? */
            {
                *ref ++ = 'F';  /* output an 'F' */
                *ref = '\0'; /* terminate the string */
                cnt ++; /* one more character in the output buffer */

                switch (inchar) /* which number? */
                {

                    case '4':
                    case '5':
                    case '6':
                    case '7':
                    case '8':
                    case '9':

                        *ref ++ = '0';  /* output an '0' */
                        *ref = '\0'; /* terminate the string */
                        cnt ++; /* one more character in the output buffer */
                        *ref ++ = (char) inchar;  /* output the number */
                        *ref = '\0'; /* terminate the string */
                        cnt ++; /* one more character in the output buffer */
                        break;

                    default:

                        *ref ++ = (char) inchar;  /* output the number */
                        *ref = '\0'; /* terminate the string */
                        cnt ++; /* one more character in the output buffer */
                        break;

                }

            }

            else
            {
                *ref ++ = (char) inchar; /* save the character in the output buffer */
                *ref = '\0'; /* terminate the string */
                cnt ++; /* one more character in the output buffer */
            }

            break;

        case '\n':

            *ref ++ = '\r'; /* save the character in the output buffer */
            *ref = '\0'; /* terminate the string */
            cnt = 0; /* new command word */
            break;

        case '\r':

            *ref ++ = (char) inchar; /* save the character in the output buffer */
            *ref = '\0'; /* terminate the string */
            cnt = 0; /* new command word */
            break;

        default:

            break;

    }

    return (local_retval);
}
